/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.sejda.sambox.pdmodel.graphics.image;

import org.sejda.sambox.util.filetypedetector.FileType;

public class UnsupportedImageFormatException extends IllegalArgumentException
{
    private final FileType fileType;
    private final String filename;

    public UnsupportedImageFormatException(FileType fileType, String filename, Throwable cause) {
        super("Image type " + fileType + " not supported " + filename, cause);
        this.filename = filename;
        this.fileType = fileType;
    }

    public FileType getFileType() {
        return fileType;
    }

    public String getFilename() {
        return filename;
    }
}
