/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.sejda.sambox.pdmodel.interactive.pagenavigation;

/**
 * The direction of motion for the specified transition effect. Only for {@link PDTransitionStyle#Split},
 * {@link PDTransitionStyle#Blinds} and {@link PDTransitionStyle#Fly}.
 * 
 * @author Andrea Vacondio
 *
 */
public enum PDTransitionMotion
{
    /**
     * Inward from the edges of the page
     */
    I,
    /**
     * Outward from the center of the page
     */
    O;
}
